from importlib import resources
from paddle import (
    setup_profile,
    write_profile,
    find_init_params,
)
from snapy import (
    MeshBlockOptions,
    MeshBlock,
)
from kintera import ThermoX
from pathlib import Path


def setup_saturn_profile():
    # path = resources.files("paddle") / "data" / "saturn1d.yaml"
    path = Path("data") / "saturn1d.yaml"
    print(f"Reading input file: {path}")

    op_block = MeshBlockOptions.from_yaml(str(path))
    block = MeshBlock(op_block)

    param = {
        "Ts": 600.0,
        "Ps": 100.0e5,
        "Tmin": 85.0,
        "xH2O": 8.91e-3,
        "xNH3": 3.52e-4,
        "xH2S": 8.08e-5,
        "grav": 10.44,
    }

    # method = "pseudo-adiabat"
    # method = "moist-adiabat"
    method = "dry-adiabat"

    param = find_init_params(
        block,
        param,
        target_T=134.0,
        target_P=1.0e5,
        method=method,
        max_iter=50,
        ftol=1.0e-2,
        verbose=False,
    )

    w = setup_profile(block, param, method=method)

    thermo_y = block.module("hydro.eos.thermo")
    thermo_x = ThermoX(thermo_y.options)
    thermo_x.to(dtype=w.dtype, device=w.device)

    write_profile("saturn_profile.txt", w, block)
    return w


if __name__ == "__main__":
    w = setup_saturn_profile()



#!/usr/bin/env python3
"""
Convert a sequence of PyTorch .pt dumps into a CF‐compliant NetCDF4 file
with dimensions (time, x, y, z) plus a 'species' axis for mole fractions.
"""

import os
import tarfile
import re
import torch
import numpy as np
from datetime import datetime
from netCDF4 import Dataset

# ──────────────────────────────── CONFIG ────────────────────────────────
INPUT_DIR = "."
OUTPUT_FILE = "thermo_x_xfrac_to_conc.nc"
# ────────────────────────────────────────────────────────────────────────

# find all .pt files, skip size==0, sort by timestamp
pt_files = []
for fn in os.listdir(INPUT_DIR):
    if not fn.endswith(".pt"):
        continue
    full = os.path.join(INPUT_DIR, fn)
    if os.path.getsize(full) == 0:
        continue
    # expect names like thermo_x_xfrac_to_conc_<epoch>.pt
    m = re.search(r"(\d+)\.pt$", fn)
    if not m:
        continue
    pt_files.append((int(m.group(1)), full))

pt_files.sort(key=lambda x: x[0])
times_epoch = [ts for ts, _ in pt_files]

# load the first file to infer shapes
module = torch.jit.load(pt_files[0][1])
data = {name: param for name, param in module.named_parameters()}

temp0 = data["temp"].numpy()
pres0 = data["pres"].numpy()
xfrac0 = data["xfrac"].numpy()
nx3, nx2, nx1 = temp0.shape
nspecies = xfrac0.shape[3]
nt = len(pt_files)

# pre‐allocate arrays in (time, x1, x2, x3) order
temp_arr = np.empty((nt, nx1, nx2, nx3), dtype=temp0.dtype)
pres_arr = np.empty((nt, nx1, nx2, nx3), dtype=pres0.dtype)
xfrac_arr = np.empty((nspecies, nt, nx1, nx2, nx3), dtype=xfrac0.dtype)

# load all timesteps
for i, (_, path) in enumerate(pt_files):
    module = torch.jit.load(path)
    data = {name: param for name, param in module.named_parameters()}
    t_np = data["temp"].numpy()  # (z, y, x)
    p_np = data["pres"].numpy()  # (z, y, x)
    x_np = data["xfrac"].numpy()  # (species, z, y, x)

    # reorder to (x, y, z)
    temp_arr[i] = t_np.transpose(2, 1, 0)
    pres_arr[i] = p_np.transpose(2, 1, 0)
    for j in range(nspecies):
        xfrac_arr[j, i] = x_np[:, :, :, j].transpose(2, 1, 0)

# create NetCDF4 file
ds = Dataset(OUTPUT_FILE, "w", format="NETCDF4")

# dimensions
ds.createDimension("time", nt)
ds.createDimension("x3", nx3)
ds.createDimension("x2", nx2)
ds.createDimension("x1", nx1)

# coordinate variables
tvar = ds.createVariable("time", "f4", ("time",))
tvar.units = "seconds since 1970-01-01 00:00:00 UTC"
tvar.calendar = "gregorian"
tvar[:] = np.array(times_epoch, dtype="f4")

zvar = ds.createVariable("x1", "f4", ("x1",))
yvar = ds.createVariable("x2", "f4", ("x2",))
xvar = ds.createVariable("x3", "f4", ("x3",))

xvar.axis = "X"
yvar.axis = "Y"
zvar.axis = "Z"

xvar[:] = np.arange(nx3)
yvar[:] = np.arange(nx2)
zvar[:] = np.arange(nx1)

# data variables
temp_v = ds.createVariable("temp", "f4", ("time", "x1", "x2", "x3"), zlib=True)
temp_v.units = "K"
temp_v.long_name = "temperature"

pres_v = ds.createVariable("pres", "f4", ("time", "x1", "x2", "x3"), zlib=True)
pres_v.units = "Pa"
pres_v.long_name = "pressure"

xfrac_v = []
for i in range(nspecies):
    xfrac_v.append(
        ds.createVariable(f"xfrac{i}", "f4", ("time", "x1", "x2", "x3"), zlib=True)
    )
    xfrac_v[i].units = "1"
    xfrac_v[i].long_name = "mole fraction of each species"

# write the data
temp_v[:] = temp_arr
pres_v[:] = pres_arr
for i in range(nspecies):
    xfrac_v[i][:] = xfrac_arr[i]

# global metadata
ds.title = "Debug fields for thermo_x.xfrac_to_conc"
ds.institution = "University of Michigan"
ds.source = "converted from .pt files"
ds.history = f"Created {datetime.utcnow().isoformat()}Z"

ds.close()
print(f"Converted file: {OUTPUT_FILE}")



import torch
import snapy
import kintera


def evolve_kinetics(
    hydro_w: torch.Tensor,
    block: snapy.MeshBlock,
    kinet: kintera.Kinetics,
    thermo_x: kintera.ThermoX,
    dt,
) -> torch.Tensor:
    """
    Evolve the chemical kinetics for one time step using implicit method.

    Args:
        hydro_w (torch.Tensor): The primitive variables tensor.
        block (snapy.MeshBlock): The mesh block containing the simulation data.
        kinet (kintera.Kinetics): The kinetics module for chemical reactions.
        thermo_x (kintera.ThermoX): The thermodynamics module for computing properties.
        dt (float): The time step for evolution.

    Returns:
        torch.Tensor: The change in mass density due to chemical reactions.
    """
    eos = block.hydro.get_eos()
    thermo_y = block.module("hydro.eos.thermo")

    temp = eos.compute("W->T", (hydro_w,))
    pres = hydro_w[snapy.index.ipr]
    xfrac = thermo_y.compute("Y->X", (hydro_w[snapy.index.icy :],))
    conc = thermo_x.compute("TPX->V", (temp, pres, xfrac))
    cp_vol = thermo_x.compute("TV->cp", (temp, conc))

    conc_kinet = kinet.options.narrow_copy(conc, thermo_y.options)
    rate, rc_ddC, rc_ddT = kinet.forward_nogil(temp, pres, conc_kinet)
    jac = kinet.jacobian(temp, conc_kinet, cp_vol, rate, rc_ddC, rc_ddT)

    stoich = kinet.buffer("stoich")
    del_conc = kintera.evolve_implicit(rate, stoich, jac, dt)

    inv_mu = thermo_y.buffer("inv_mu")
    del_rho = del_conc / inv_mu[1:].view(1, 1, 1, -1)
    return del_rho.permute(3, 0, 1, 2)



from typing import Tuple

import torch
import snapy
import kintera


def integrate_neutral(
    thermo_x: kintera.ThermoX,
    temp: torch.Tensor,
    pres: torch.Tensor,
    xfrac: torch.Tensor,
    grav: float,
    dz: float,
    max_iter: int = 100,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    A neutral density profile assumes no cloud and:

        (1) dP/dz = -rho*g
        (2) d(rho)/dz = ...

    In discretized form:

        rho_bar = 0.5 * (rho_old + rho_ad)
        P_new = P_old - rho_bar * g * dz
    """
    conc = thermo_x.compute("TPX->V", [temp, pres, xfrac])
    rho = thermo_x.compute("V->D", [conc])

    # make an adiabatic step first
    temp_ad = temp.clone()
    pres_ad = pres.clone()
    xfrac_ad = xfrac.clone()

    thermo_x.extrapolate_ad(temp_ad, pres_ad, xfrac_ad, grav, dz)
    conc_ad = thermo_x.compute("TPX->V", [temp_ad, pres_ad, xfrac_ad])
    rho_ad = thermo_x.compute("V->D", [conc_ad])
    rho_bar = 0.5 * (rho + rho_ad)

    pres2 = pres - rho_bar * grav * dz

    # initial guess
    temp2 = temp_ad.clone()
    count = 0
    while count < max_iter:
        xfrac2 = xfrac_ad.clone()

        # equilibrate clouds
        thermo_x.forward(temp2, pres2, xfrac2)

        # drop clouds fractions
        for cid in thermo_x.options.cloud_ids():
            xfrac2[..., cid] = 0.0
        # renormalize mole fractions
        xfrac2 /= xfrac.sum(dim=-1, keepdim=True)

        conc2 = thermo_x.compute("TPX->V", [temp2, pres2, xfrac2])
        rho2 = thermo_x.compute("V->D", [conc2])

        if torch.allclose(rho2, rho_ad):
            break

        temp2 -= temp2 * (rho_ad - rho2) / rho2
        count += 1

    if count == max_iter:
        raise RuntimeError("neutral density integration did not converge.")

    return temp2, pres2, xfrac2


def integrate_dry_adiabat(
    thermo_x: kintera.ThermoX,
    temp: torch.Tensor,
    pres: torch.Tensor,
    xfrac: torch.Tensor,
    grav: float,
    dz: float,
    max_iter: int = 100,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    A dry adiabatic profile assumes no cloud and:

        (1) dT/dz = -g/cp
        (2) dP/dz = -rho*g

    In discretized form:

        cp_bar = 0.5 * (cp(T_old) + cp(T_new))
        T_new = T_old - g/bar * dz
        rho_bar = 0.5 * (rho_old + rho_new)
        P_new = P_old - rho_bar * g * dz

    """
    conc1 = thermo_x.compute("TPX->V", [temp, pres, xfrac])
    cp1 = thermo_x.compute("TV->cp", [temp, conc1]) / conc1.sum(-1)
    rho1 = thermo_x.compute("V->D", [conc1])
    mmw1 = (thermo_x.mu * xfrac).sum(-1)

    # initial guess
    temp2 = temp - grav * mmw1 * dz / cp1
    pres2 = pres - rho1 * grav * dz

    count = 0
    while count < max_iter:
        xfrac2 = xfrac.clone()

        # equilibrate clouds
        thermo_x.forward(temp2, pres2, xfrac2)

        # drop clouds fractions
        for cid in thermo_x.options.cloud_ids():
            xfrac2[..., cid] = 0.0
        # renormalize mole fractions
        xfrac2 /= xfrac.sum(dim=-1, keepdim=True)

        conc2 = thermo_x.compute("TPX->V", [temp2, pres2, xfrac2])
        cp2 = thermo_x.compute("TV->cp", [temp2, conc2]) / conc2.sum(-1)
        rho2 = thermo_x.compute("V->D", [conc2])
        mmw2 = (thermo_x.mu * xfrac2).sum(-1)

        cp_bar = 0.5 * (cp1 / mmw1 + cp2 / mmw2)
        rho_bar = 0.5 * (rho1 + rho2)

        temp_new = temp - grav * dz / cp_bar
        pres_new = pres - rho_bar * grav * dz

        if torch.allclose(temp_new, temp2) and torch.allclose(pres_new, pres2):
            break

        temp2 = temp_new
        pres2 = pres_new
        count += 1

    if count == max_iter:
        raise RuntimeError("Dry adiabat integration did not converge.")

    return temp2, pres2, xfrac2


def setup_profile(
    block: snapy.MeshBlock,
    param: dict[str, float] = {},
    method: str = "moist-adiabat",
    verbose: bool = False,
) -> torch.Tensor:
    """
    Set up an adiabatic initial condition for the mesh block.

    This function initializes the primitive variables in the mesh block
    and returns the initialized tensor.

    Args:
        block (snapy.MeshBlock): The mesh block to set up.
        param (dict[str, float], optional): Parameters for the adiabat setup. Defaults to {}.
        method (str, optional): Method for the adiabat setup. Choose between
            (1) "dry-adiabat"
            (2) "moist-adiabat"
            (3) "isothermal"
            (4) "pseudo-adiabat"
            (5) "neutral"
            Defaults to "moist-adiabat".

        Required parameters in `param`:
            Ts (float): Surface temperature in Kelvin. Default is 300 K.
            Ps (float): Surface pressure in Pascals. Default is 1e5 Pa.
            x<species> (float): Mole fraction of a specific species (e.g., xH2O for
            water vapor). Default is 0.0.
            grav (float): Gravitational acceleration in m/s^2. Default is 9.8 m/s^2.

    Returns:
        torch.Tensor: The initialized primitive variables tensor.
    """

    # check method
    valid_methods = [
        "dry-adiabat",
        "moist-adiabat",
        "isothermal",
        "pseudo-adiabat",
        "neutral",
    ]

    if method not in valid_methods:
        raise ValueError(f"Invalid method '{method}'. Choose from {valid_methods}.")

    Ts = param.get("Ts", 300.0)
    Ps = param.get("Ps", 1.0e5)
    grav = param.get("grav", 9.8)
    Tmin = param.get("Tmin", 0.0)

    # get handles to modules
    coord = block.module("hydro.coord")
    thermo_y = block.module("hydro.eos.thermo")

    # get coordinates
    x3v, x2v, x1v = torch.meshgrid(
        coord.buffer("x3v"), coord.buffer("x2v"), coord.buffer("x1v"), indexing="ij"
    )

    # handling mole fractions
    thermo_x = kintera.ThermoX(thermo_y.options)
    thermo_x.to(dtype=x1v.dtype, device=x1v.device)

    # get dimensions
    nc3, nc2, nc1 = x1v.shape
    ny = len(thermo_y.options.species()) - 1
    nvar = 5 + ny

    w = torch.zeros((nvar, nc3, nc2, nc1), dtype=x1v.dtype, device=x1v.device)

    temp = Ts * torch.ones((nc3, nc2), dtype=w.dtype, device=w.device)
    pres = Ps * torch.ones((nc3, nc2), dtype=w.dtype, device=w.device)
    xfrac = torch.zeros((nc3, nc2, ny + 1), dtype=w.dtype, device=w.device)

    for name in thermo_y.options.species():
        index = thermo_y.options.species().index(name)
        xfrac[..., index] = param.get(f"x{name}", 0.0)

    # dry air mole fraction
    xfrac[..., 0] = 1.0 - xfrac[..., 1:].sum(dim=-1)

    # start and end indices for the vertical direction
    # excluding ghost cells
    ifirst = coord.ifirst()
    ilast = coord.ilast()

    # vertical grid distance of the first cell
    dz = coord.buffer("dx1f")[ifirst]

    # half a grid to cell center
    thermo_x.extrapolate_ad(temp, pres, xfrac, grav, dz / 2.0, verbose=verbose)

    # adiabatic extrapolation
    if method == "isothermal":
        i_isothermal = ifirst
        ifirst = ilast
    else:
        i_isothermal = ilast

    for i in range(ifirst, ilast):
        # drop clouds fractions
        if method.split("-")[0] != "moist":
            for cid in thermo_x.options.cloud_ids():
                xfrac[..., cid] = 0.0
            # renormalize mole fractions
            xfrac /= xfrac.sum(dim=-1, keepdim=True)
        conc = thermo_x.compute("TPX->V", [temp, pres, xfrac])

        w[snapy.index.ipr, ..., i] = pres
        w[snapy.index.idn, ..., i] = thermo_x.compute("V->D", [conc])
        w[snapy.index.icy :, ..., i] = thermo_x.compute("X->Y", [xfrac])

        dz = coord.buffer("dx1f")[i]
        if method.split("-")[0] == "dry":
            temp, pres, xfrac = integrate_dry_adiabat(
                thermo_x, temp, pres, xfrac, grav, dz
            )
        elif method.split("-")[0] == "neutral":
            temp, pres, xfrac = integrate_neutral(thermo_x, temp, pres, xfrac, grav, dz)
        else:
            thermo_x.extrapolate_ad(temp, pres, xfrac, grav, dz, verbose=verbose)

        if torch.any(temp < Tmin):
            i_isothermal = i + 1
            break

    # isothermal extrapolation
    for i in range(i_isothermal, ilast):
        # drop clouds fractions
        if method.split("-")[0] != "moist":
            for cid in thermo_x.options.cloud_ids():
                xfrac[..., cid] = 0.0
            # renormalize mole fractions
            xfrac /= xfrac.sum(dim=-1, keepdim=True)

        mu = (thermo_x.mu * xfrac).sum(-1)
        dz = coord.buffer("dx1f")[i]
        pres *= torch.exp(-grav * mu * dz / (kintera.constants.Rgas * temp))
        conc = thermo_x.compute("TPX->V", [temp, pres, xfrac])
        w[snapy.index.ipr, ..., i] = pres
        w[snapy.index.idn, ..., i] = thermo_x.compute("V->D", [conc])
        w[snapy.index.icy :, ..., i] = thermo_x.compute("X->Y", [xfrac])
    return w



import torch
import math
import time
from snapy import MeshBlockOptions, MeshBlock
from snapy import kIDN, kIPR


def call_user_output(bvars):
    hydro_w = bvars["hydro_w"]
    out = {}
    temp = hydro_w[kIPR] / (Rd * hydro_w[kIDN])
    out["temp"] = temp
    out["theta"] = temp * (p0 / hydro_w[kIPR]).pow(Rd / cp)
    return out


dT = 0.5
p0 = 1.0e5
Ts = 303.15
xc = 500.0
yc = 0.0
zc = 260.0
s = 100.0
a = 50.0
grav = 9.8
Rd = 287.0
gamma = 1.4
uniform_bubble = False

# use cuda if available
if torch.cuda.is_available():
    device = torch.device("cuda:0")
else:
    device = torch.device("cpu")

# set hydrodynamic options
op = MeshBlockOptions.from_yaml("robert.yaml")
block = MeshBlock(op)
block.to(device)

# get handles to modules
coord = block.module("hydro.coord")
thermo = block.module("hydro.eos.thermo")

# thermodynamics
cp = gamma / (gamma - 1.0) * Rd

# set initial condition
x3v, x2v, x1v = torch.meshgrid(
    coord.buffer("x3v"), coord.buffer("x2v"), coord.buffer("x1v"), indexing="ij"
)

# dimensions
nc3 = coord.buffer("x3v").shape[0]
nc2 = coord.buffer("x2v").shape[0]
nc1 = coord.buffer("x1v").shape[0]
nvar = 5

w = torch.zeros((nvar, nc3, nc2, nc1), device=device)

temp = Ts - grav * x1v / cp
w[kIPR] = p0 * torch.pow(temp / Ts, cp / Rd)

r = torch.sqrt((x3v - yc) ** 2 + (x2v - xc) ** 2 + (x1v - zc) ** 2)
temp += torch.where(r <= a, dT * torch.pow(w[kIPR] / p0, Rd / cp), 0.0)
if not uniform_bubble:
    temp += torch.where(
        r > a,
        dT * torch.exp(-(((r - a) / s) ** 2)) * torch.pow(w[kIPR] / p0, Rd / cp),
        0.0,
    )
w[kIDN] = w[kIPR] / (Rd * temp)

block_vars = {}
block_vars["hydro_w"] = w
block_vars, current_time = block.initialize(block_vars)

block.set_user_output_func(call_user_output)

# integration
start_time = time.time()
block.make_outputs(block_vars, current_time)

while not block.intg.stop(block.inc_cycle(), current_time):
    dt = block.max_time_step(block_vars)
    block.print_cycle_info(block_vars, current_time, dt)

    for stage in range(len(block.intg.stages)):
        block.forward(block_vars, dt, stage)

    err = block.check_redo(block_vars)
    if err > 0:
        continue  # redo current step
    if err < 0:
        break  # terminate

    current_time += dt
    block.make_outputs(block_vars, current_time)

block.finalize(block_vars, current_time)
